import asyncio
from concurrent.futures import Future
from concurrent.futures import ProcessPoolExecutor as ProcessPoolExecutorBase
from multiprocessing import cpu_count
from typing import Any

from .compat import EventLoopMixin
from .counters import Statistic


class ProcessPoolStatistic(Statistic):
    processes: int
    done: int
    error: int
    success: int
    submitted: int
    sum_time: float


class ProcessPoolExecutor(ProcessPoolExecutorBase, EventLoopMixin):
    """
    Process pool executor with statistic

    Usage:

    .. code-block:: python

        from time import sleep
        from aiomisc import ProcessPoolExecutor

        # NOTE: blocking function must be defined at the top level
        # of the module to be able to be pickled and sent to the
        # child processes.
        def blocking_fn():
            sleep(1)
            return 42

        async def main():
            executor = ProcessPoolExecutor()
            the_answer = await executor.submit(blocking_fn)
            print("The answer is:", the_answer)

        asyncio.run(main())
    """

    DEFAULT_MAX_WORKERS = max((cpu_count(), 4))

    def __init__(self, max_workers: int = DEFAULT_MAX_WORKERS, **kwargs: Any):
        """
        Initializes a new ProcessPoolExecutor instance.

        * ``max_workers``:
          The maximum number of processes that can be used to
          execute the given calls. If None or not given then
          as many worker processes will be created as the
          machine has processors.
        * ``mp_context``:
          A multiprocessing context to launch the workers. This
          object should provide SimpleQueue, Queue and Process.
          Useful to allow specific multiprocessing start methods.
        * ``initializer``:
          A callable used to initialize worker processes.
        * ``initargs``:
          A tuple of arguments to pass to the initializer.
        * ``max_tasks_per_child``:
          The maximum number of tasks a worker process
          can complete before it will exit and be replaced
          with a fresh worker process. The default of None
          means worker process will live as long as the
          executor. Requires a non-'fork' mp_context start
          method. When given, we default to using 'spawn'
          if no mp_context is supplied.
        """
        super().__init__(max_workers=max_workers, **kwargs)
        self._statistic = ProcessPoolStatistic()
        self._statistic.processes = max_workers

    def _statistic_callback(
        self,
        future: Future,
        start_time: float,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        """
        Callback for statistic
        """
        if future.exception():
            self._statistic.error += 1
        else:
            self._statistic.success += 1
        self._statistic.done += 1
        self._statistic.sum_time += loop.time() - start_time

    def submit(self, *args: Any, **kwargs: Any) -> Future:
        """
        Submit blocking function to the pool
        """
        loop = asyncio.get_running_loop()
        start_time = loop.time()
        future = super().submit(*args, **kwargs)
        self._statistic.submitted += 1
        future.add_done_callback(
            lambda f: self._statistic_callback(f, start_time, loop),
        )
        return future

    def __del__(self) -> None:
        """
        Cleanup resources
        """
        self.shutdown()



import atexit
import os
import sys
from contextlib import suppress
from subprocess import PIPE, Popen
from threading import Event
from time import sleep
from typing import Any, Callable, Mapping, MutableMapping, Tuple

from aiomisc_log import basic_config

from . import INT_SIGNAL, AddressType, log
from .protocol import FileIOProtocol


STOPPING = Event()


class Worker:
    def __init__(
        self, log_level: str, log_format: str, address: AddressType,
        cookie: bytes, worker_id: bytes, env: Mapping[str, str],
        initializer: Callable[..., Any],
        initializer_args: Any, initializer_kwargs: Any,
    ):
        env = dict(env)
        self.process = Popen(
            [sys.executable, "-m", "aiomisc_worker.process_inner"],
            stdin=PIPE, env=env,
        )

        assert self.process.stdin

        proto = FileIOProtocol(self.process.stdin)
        proto.send((log_level, log_format))
        proto.send(address)
        proto.send(cookie)
        proto.send(worker_id)
        proto.send((initializer, initializer_args, initializer_kwargs))
        self.process.stdin.close()
        atexit.register(self.close)

    @property
    def is_running(self) -> bool:
        return self.process.poll() is None

    def kill(self, sig: int = INT_SIGNAL) -> None:
        if not self.is_running:
            return
        os.kill(self.process.pid, sig)

    def close(self) -> None:
        with suppress(Exception):
            self.kill()
        atexit.unregister(self.close)

    def __del__(self) -> None:
        self.close()


PROCESSES: MutableMapping[Worker, bytes] = {}


def at_exit() -> None:
    processes = tuple(PROCESSES.keys())
    PROCESSES.clear()
    for process in processes:
        process.close()


def main() -> int:
    proto = FileIOProtocol(sys.stdin.buffer)

    log_level, log_format = proto.receive()
    basic_config(level=log_level, log_format=log_format)

    address: AddressType = proto.receive()
    cookie: bytes = proto.receive()
    worker_ids: Tuple[bytes, ...] = proto.receive()
    initializer, initializer_args, initializer_kwargs = proto.receive()

    sys.stdin.close()
    del proto

    env = dict(os.environ)
    env["AIOMISC_NO_PLUGINS"] = ""

    def create_worker() -> Worker:
        nonlocal env        # noqa
        return Worker(
            log_level, log_format, address, cookie, worker_id, env,
            initializer, initializer_args, initializer_kwargs,
        )

    log.debug("Starting %d processes", len(worker_ids))
    for worker_id in worker_ids:
        worker = create_worker()
        PROCESSES[worker] = worker_id

    log.info("Waiting workers")

    atexit.register(at_exit)

    try:
        while True:
            for worker in tuple(PROCESSES.keys()):
                if worker.is_running:
                    continue

                worker.close()
                log.debug(
                    "Worker PID: %d exited with status %d",
                    worker.process.pid, worker.process.returncode,
                )
                worker_id = PROCESSES.pop(worker)
                worker = create_worker()
                PROCESSES[worker] = worker_id
            sleep(0.01)
    except KeyboardInterrupt:
        pass

    return 0



import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Hashable, Iterable, List

import dnslib  # type: ignore[import-untyped]


class RecordType(enum.IntEnum):
    A = 1
    A6 = 38
    AAAA = 28
    AFSDB = 18
    ANY = 255
    APL = 42
    AXFR = 252
    CAA = 257
    CDNSKEY = 60
    CDS = 59
    CERT = 37
    CNAME = 5
    CSYNC = 62
    DHCID = 49
    DLV = 32769
    DNAME = 39
    DNSKEY = 48
    DS = 43
    EUI48 = 108
    EUI64 = 109
    HINFO = 13
    HIP = 53
    HIP_AC = 55
    HTTPS = 65
    IPSECKEY = 45
    IXFR = 251
    KEY = 25
    KX = 36
    LOC = 29
    MX = 15
    NAPTR = 35
    NS = 2
    NSEC = 47
    NSEC3 = 50
    NSEC3PARAM = 51
    NULL = 10
    OPENPGPKEY = 61
    OPT = 41
    PTR = 12
    RP = 17
    RRSIG = 46
    SIG = 24
    SOA = 6
    SPF = 99
    SRV = 33
    SSHFP = 44
    SVCB = 64
    TA = 32768
    TKEY = 249
    TLSA = 52
    TSIG = 250
    TXT = 16
    URI = 256
    ZONEMD = 63


class DNSClass(enum.IntEnum):
    IN = 1
    CS = 2
    CH = 3
    Hesiod = 4
    NONE = 254
    ASTERISK = 255


class RD(dnslib.RD, Hashable, ABC):
    def __hash__(self) -> int:
        return hash(self.data)

    @classmethod
    @abstractmethod
    def create(cls, *args: Any, **kwargs: Any) -> "DNSRecord":
        raise NotImplementedError


@dataclass(frozen=True)
class DNSRecord:
    name: str
    type: RecordType
    data: RD
    cls: DNSClass = field(default=DNSClass.IN)
    ttl: int = field(default=3600, compare=False)

    def rr(self, query_type: int) -> dnslib.RR:
        return dnslib.RR(
            rname=self.name,
            rtype=query_type,
            rclass=self.cls,
            ttl=self.ttl,
            rdata=self.data,
        )


class A(dnslib.A, RD):
    @classmethod
    def create(cls, name: str, ip: str, ttl: int = 3600) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.A,
            data=cls(ip),
            ttl=ttl,
        )


class AAAA(dnslib.AAAA, RD):
    @classmethod
    def create(cls, name: str, ipv6: str, ttl: int = 3600) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.AAAA,
            data=cls(ipv6),
            ttl=ttl,
        )


class RDLabel(RD):
    label: bytes
    __type__: RecordType

    @classmethod
    def create(cls, name: str, label: str, ttl: int = 3600) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=cls.__type__,
            data=cls(label),
            ttl=ttl,
        )

    def __hash__(self) -> int:
        return hash(self.label)


class CNAME(dnslib.CNAME, RDLabel):
    __type__ = RecordType.CNAME


class NS(dnslib.NS, RDLabel):
    __type__ = RecordType.NS


class PTR(dnslib.PTR, RDLabel):
    __type__ = RecordType.PTR


class DNAME(dnslib.DNAME, RDLabel):
    __type__ = RecordType.DNAME


class MX(dnslib.MX, RD):
    @classmethod
    def create(
        cls, name: str, exchange: str, preference: int, ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.MX,
            data=cls(exchange, preference),
            ttl=ttl,
        )


class TXT(dnslib.TXT, RD):
    @classmethod
    def create(cls, name: str, text: str, ttl: int = 3600) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.TXT,
            data=cls(text),
            ttl=ttl,
        )


class SOA(dnslib.SOA, RD):
    @classmethod
    def create(
        cls, name: str, mname: str, rname: str, serial: int,
        refresh: int, retry: int, expire: int, minimum: int,
        ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.SOA,
            data=cls(
                mname, rname,
                (serial, refresh, retry, expire, minimum),
            ),
            ttl=ttl,
        )


class SRV(dnslib.SRV, RD):
    @classmethod
    def create(
        cls, name: str, priority: int, weight: int, port: int,
        target: str, ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.SRV,
            data=cls(priority, weight, port, target),
            ttl=ttl,
        )


class CAA(dnslib.CAA, RD):
    @classmethod
    def create(
        cls, name: str, flags: int, tag: str, value: str,
        ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.CAA,
            data=cls(flags, tag, value),
            ttl=ttl,
        )


class NAPTR(dnslib.NAPTR, RD):
    @classmethod
    def create(
        cls, name: str, order: int, preference: int, flags: str,
        service: str, regexp: str, replacement: str, ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.NAPTR,
            data=cls(order, preference, flags, service, regexp, replacement),
            ttl=ttl,
        )


class DS(dnslib.DS, RD):
    @classmethod
    def create(
        cls, name: str, key_tag: int, algorithm: int, digest_type: int,
        digest: str, ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.DS,
            data=cls(key_tag, algorithm, digest_type, digest),
            ttl=ttl,
        )


class DNSKEY(dnslib.DNSKEY, RD):
    @classmethod
    def create(
        cls, name: str, flags: int, protocol: int, algorithm: int,
        key: str, ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.DNSKEY,
            data=cls(flags, protocol, algorithm, key),
            ttl=ttl,
        )


class RRSIG(dnslib.RRSIG, RD):
    @classmethod
    def create(
        cls, name: str, type_covered: int, algorithm: int, labels: int,
        original_ttl: int, expiration: int, inception: int, key_tag: int,
        signer: str, signature: str, ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.RRSIG,
            data=cls(
                type_covered, algorithm, labels, original_ttl, expiration,
                inception, key_tag, signer, signature,
            ),
            ttl=ttl,
        )


class NSEC(dnslib.NSEC, RD):
    @classmethod
    def create(
        cls, name: str, next_domain: str,
        rrtypes: Iterable[int], ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.NSEC,
            data=cls(next_domain, list(rrtypes)),
            ttl=ttl,
        )


class HTTPS(dnslib.HTTPS, RD):
    @classmethod
    def create(
        cls, name: str, priority: int, target: str,
        params: List[str], ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.HTTPS,
            data=cls(priority, target, params),
            ttl=ttl,
        )


class LOC(dnslib.LOC, RD):
    @classmethod
    def create(
        cls, name: str, latitude: float, longitude: float,
        altitude: float, size: float, h_precision: float,
        v_precision: float, ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.LOC,
            data=cls(
                latitude, longitude, altitude, size, h_precision, v_precision,
            ),
            ttl=ttl,
        )


class RP(dnslib.RP, RD):
    @classmethod
    def create(
        cls, name: str, mbox: str, txt: str, ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.RP,
            data=cls(mbox, txt),
            ttl=ttl,
        )


class TLSA(dnslib.TLSA, RD):
    @classmethod
    def create(
        cls, name: str, usage: int, selector: int, mtype: int, cert: str,
        ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.TLSA,
            data=cls(usage, selector, mtype, cert),
            ttl=ttl,
        )


class SSHFP(dnslib.SSHFP, RD):
    @classmethod
    def create(
        cls, name: str, algorithm: int, fptype: int,
        fingerprint: str, ttl: int = 3600,
    ) -> DNSRecord:
        return DNSRecord(
            name=name,
            type=RecordType.SSHFP,
            data=cls(algorithm, fptype, fingerprint),
            ttl=ttl,
        )


__all__ = (
    "AAAA",
    "CAA",
    "CNAME",
    "DNAME",
    "DNSKEY",
    "DS",
    "HTTPS",
    "LOC",
    "MX",
    "NAPTR",
    "NS",
    "NSEC",
    "PTR",
    "RD",
    "RP",
    "RRSIG",
    "SOA",
    "SRV",
    "SSHFP",
    "TLSA",
    "TXT",
    "A",
    "DNSRecord",
    "DNSClass",
    "RecordType",
)



import asyncio
import logging.handlers
import threading
import traceback
import warnings
from contextlib import suppress
from queue import Empty, Queue
from socket import socket
from typing import (
    Any, Callable, Dict, Iterable, List, Optional, Tuple, Type, Union,
)

import aiomisc_log
from aiomisc_log.enum import LogFormat, LogLevel

from .counters import Statistic


class ThreadedHandlerStatistic(Statistic):
    threads: int
    records: int
    errors: int
    flushes: int


class ThreadedHandler(logging.Handler):
    def __init__(
        self, target: logging.Handler, flush_interval: float = 0.1,
        buffered: bool = True, queue_size: int = 0,
    ):
        super().__init__()
        self._buffered = buffered
        self._target = target
        self._flush_interval = flush_interval
        self._flush_event = threading.Event()
        self._queue: Queue[Optional[logging.LogRecord]] = Queue(queue_size)
        self._close_event = threading.Event()
        self._thread = threading.Thread(target=self._in_thread, daemon=True)
        self._statistic = ThreadedHandlerStatistic()

    def start(self) -> None:
        self._statistic.threads += 1
        self._thread.start()

    def close(self) -> None:
        self._queue.put(None)
        del self._queue
        self.flush()
        self._close_event.set()
        super().close()

    def flush(self) -> None:
        self._statistic.flushes += 1
        self._flush_event.set()

    def emit(self, record: logging.LogRecord) -> None:
        if self._buffered:
            self._queue.put_nowait(record)
        else:
            self._queue.put(record)
        self._statistic.records += 1

    def _in_thread(self) -> None:
        queue = self._queue
        while not self._close_event.is_set():
            self._flush_event.wait(self._flush_interval)
            try:
                self.acquire()
                while True:
                    record = queue.get(timeout=self._flush_interval)
                    if record is None:
                        return
                    with suppress(Exception):
                        self._target.handle(record)
            except Empty:
                pass
            finally:
                self.release()
        self._statistic.threads -= 1


def suppressor(
    callback: Callable[..., None],
    exceptions: Tuple[Type[BaseException], ...] = (Exception,),
) -> Callable[..., None]:
    def wrapper() -> None:
        with suppress(*exceptions):
            callback()
    return wrapper


def wrap_logging_handler(
    handler: logging.Handler,
    buffer_size: int = 1024,
    flush_interval: Union[float, int] = 0.1,
    loop: Optional[asyncio.AbstractEventLoop] = None,
) -> logging.Handler:
    warnings.warn("wrap_logging_handler is deprecated", DeprecationWarning)
    handler = ThreadedHandler(
        target=handler,
        queue_size=buffer_size,
        flush_interval=flush_interval,
    )
    handler.start()
    return handler


class UnhandledLoopHook(aiomisc_log.UnhandledHook):
    @staticmethod
    def _fill_transport_extra(
        transport: Optional[asyncio.Transport],
        extra: Dict[str, Any],
    ) -> None:
        if transport is None:
            return

        extra["transport"] = repr(transport)

        for key in (
            "peername", "sockname", "compression",
            "cipher", "peercert", "pipe", "subprocess",
        ):
            value = transport.get_extra_info(key)
            if value:
                extra[f"transport_{key}"] = value

    def __call__(
        self, loop: asyncio.AbstractEventLoop,
        context: Dict[str, Any],
    ) -> None:
        context = dict(context)
        message: str = context.pop("message", "unhandled loop exception")
        exception: Optional[BaseException] = context.pop("exception", None)
        future: Optional[asyncio.Future] = context.pop("future", None)
        task: Optional[asyncio.Task] = context.pop("task", None)
        handle: Optional[asyncio.Handle] = context.pop("handle", None)
        protocol: Optional[asyncio.Protocol] = context.pop("protocol", None)
        transport: Optional[asyncio.Transport] = context.pop("transport", None)
        sock: Optional[socket] = context.pop("socket", None)
        source_tb: List[traceback.FrameSummary] = (
            context.pop("source_traceback", None) or []
        )

        if exception is None:
            if future is not None:
                exception = future.exception()
            elif task is not None and task.done():
                exception = task.exception()

        extra = context
        if handle is not None:
            extra["handle"] = repr(handle)
        if protocol is not None:
            extra["protocol"] = repr(protocol)
        if sock is not None:
            extra["sock"] = repr(sock)

        self._fill_transport_extra(transport, extra)
        self.logger.exception(message, exc_info=exception, extra=extra)
        if source_tb:
            self.logger.error("".join(traceback.format_list(source_tb)))


def basic_config(
    level: Union[int, str] = LogLevel.default(),
    log_format: Union[str, LogFormat] = LogFormat.default(),
    buffered: bool = True,
    buffer_size: int = 0,
    flush_interval: Union[int, float] = 0.2,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    handlers: Iterable[logging.Handler] = (),
    **kwargs: Any,
) -> None:
    unhandled_hook = UnhandledLoopHook(logger_name="asyncio.unhandled")

    if loop is None:
        loop = asyncio.get_event_loop()

    forever_task = asyncio.gather(
        loop.create_future(), return_exceptions=True,
    )
    loop.set_exception_handler(unhandled_hook)

    log_handlers = []

    for user_handler in handlers:
        handler = ThreadedHandler(
            buffered=buffered,
            flush_interval=flush_interval,
            queue_size=buffer_size,
            target=user_handler,
        )
        unhandled_hook.add_handler(handler)
        forever_task.add_done_callback(lambda _: handler.close())
        log_handlers.append(handler)
        handler.start()

    aiomisc_log.basic_config(
        level=level, log_format=log_format, handlers=log_handlers, **kwargs,
    )


__all__ = (
    "LogFormat",
    "LogLevel",
    "basic_config",
    "ThreadedHandler",
)



import logging
from abc import abstractmethod
from typing import Any, Optional, Union

from aiomisc import PeriodicCallback, Service


log = logging.getLogger(__name__)


class PeriodicService(Service):

    __required__ = ("interval",)

    interval: Union[int, float]
    delay: Union[int, float] = 0

    def __init__(self, *, name: Optional[str] = None, **kwargs: Any):
        super().__init__(**kwargs)
        self.periodic = PeriodicCallback(self.callback, name=name)

    async def start(self) -> None:
        assert self.interval, f"Interval illegal interval {self.interval!r}"
        assert self.interval > 0, (
            f"Interval must be positive not {self.interval!r}"
        )

        self.periodic.start(self.interval, delay=self.delay, loop=self.loop)
        log.info("Periodic service %s started", self)

    async def stop(self, err: Optional[Exception] = None) -> None:
        await self.periodic.stop(return_exceptions=True)
        log.info("Periodic service %s is stopped", self)

    @abstractmethod
    async def callback(self) -> Any:
        raise NotImplementedError

    def __str__(self) -> str:
        return "{}(interval={},delay={})".format(
            self.__class__.__name__,
            self.interval,
            self.delay,
        )



import asyncio
import logging
from asyncio import StreamReader, StreamWriter
from struct import Struct
from typing import Optional, Tuple

import dnslib  # type: ignore[import-untyped]

from aiomisc.service import TCPServer, UDPServer

from .records import RecordType
from .store import DNSStore


log = logging.getLogger(__name__)


class DNSServer:
    store: DNSStore

    __proto__: str

    @staticmethod
    def get_edns_max_size(record: dnslib.DNSRecord) -> Optional[int]:
        for rr in record.ar:
            if rr.rtype != dnslib.QTYPE.OPT:
                continue
            return rr.edns_len
        return None

    def handle_request(
        self, addr: tuple, data: bytes, should_truncate: bool = False,
    ) -> Tuple[bytes, tuple]:
        record = dnslib.DNSRecord.parse(data)
        question: dnslib.DNSQuestion = record.get_q()
        reply = record.reply()
        query_name = str(question.get_qname())
        query_type = question.qtype

        try:
            log.debug(
                "Processing %s request from %r:\n%s\n",
                self.__proto__, addr, record,
            )
            edns_max_size = self.get_edns_max_size(record)
            if edns_max_size is not None:
                reply.add_ar(dnslib.EDNS0(udp_len=edns_max_size))

            records = self.store.query(query_name, RecordType(query_type))
            for rec in records:
                reply.add_answer(rec.rr(query_type))

            if not records:
                reply.header.rcode = dnslib.RCODE.NXDOMAIN

            log.debug(
                "Sending %s answer to %r:\n%s\n",
                self.__proto__, addr, reply,
            )

            reply_body = reply.pack()

            if len(reply_body) > 512:
                reply.header.tc = 1
                reply_body = reply.pack()

            if should_truncate:
                reply_body = reply_body[:edns_max_size]
        except Exception as e:
            log.exception(
                "Failed to process %s request from %r: %s",
                self.__proto__, addr, e,
            )
            reply = record.reply()
            reply.header.set_rcode(dnslib.RCODE.SERVFAIL)
            reply_body = reply.pack()

        return reply_body, addr


class UDPDNSServer(DNSServer, UDPServer):
    MAX_SIZE = 512

    store: DNSStore

    __required__ = ("store",)
    __proto__ = "UDP"

    async def handle_datagram(self, data: bytes, addr: tuple) -> None:
        self.sendto(
            *self.handle_request(
                data=data, addr=addr, should_truncate=True,
            ),
        )


TCP_HEADER_STRUCT: Struct = Struct("!H")


class TCPDNSServer(DNSServer, TCPServer):
    store: DNSStore

    __required__ = ("store",)

    __proto__ = "TCP"

    async def handle_client(
        self, reader: StreamReader, writer: StreamWriter,
    ) -> None:
        addr = writer.get_extra_info("peername")
        try:
            while True:
                data = await reader.readexactly(TCP_HEADER_STRUCT.size)
                if not data:
                    break
                length = TCP_HEADER_STRUCT.unpack(data)[0]
                packet = await reader.read(length)

                if not data:
                    break

                reply_body, _ = self.handle_request(data=packet, addr=addr)

                writer.write(TCP_HEADER_STRUCT.pack(len(reply_body)))
                writer.write(reply_body)
                await writer.drain()
        except asyncio.IncompleteReadError:
            pass
        finally:
            writer.close()
            await writer.wait_closed()



from typing import Iterable, Mapping, Optional, Sequence, Tuple

from .records import DNSRecord, RecordType
from .tree import RadixTree
from .zone import DNSZone


class DNSStore:
    zones: RadixTree[DNSZone]

    __slots__ = ("zones",)

    def __init__(self) -> None:
        self.zones = RadixTree()

    def add_zone(self, zone: DNSZone) -> None:
        zone_tuple = self.get_reverse_tuple(zone.name)
        if self.zones.search(zone_tuple):
            raise ValueError(f"Zone {zone.name} already exists.")
        self.zones.insert(zone_tuple, zone)

    def remove_zone(self, zone_name: str) -> None:
        zone_tuple = self.get_reverse_tuple(zone_name)
        if not self.zones.search(zone_tuple):
            raise ValueError(
                f"Zone {zone_name} does not exist.",
            )
        # Clear zone from RadixTree
        self.zones.insert(zone_tuple, None)

    def get_zone(self, zone_name: str) -> Optional[DNSZone]:
        zone_tuple = self.get_reverse_tuple(zone_name)
        return self.zones.search(zone_tuple)

    def query(self, name: str, record_type: RecordType) -> Sequence[DNSRecord]:
        if not name.endswith("."):
            name += "."
        zone_tuple = self.get_zone_for_name(name)
        if not zone_tuple:
            return ()
        zone = self.zones.search(zone_tuple)
        return zone.get_records(name, record_type) if zone is not None else ()

    def get_zone_for_name(self, name: str) -> Optional[Tuple[str, ...]]:
        labels = self.get_reverse_tuple(name)
        result = self.zones.find_prefix(labels)
        return result[0] if result else None

    @staticmethod
    def get_reverse_tuple(zone_name: str) -> Tuple[str, ...]:
        return tuple(zone_name.strip(".").split("."))[::-1]

    def replace(
        self, zones_data: Mapping[str, Iterable[DNSRecord]],
    ) -> None:
        """
        Atomically replace all zones with new ones this method is safe
        because it replaces all zones at once. zone_data is a mapping
        zone name and a sequence of DNSRecord objects.

        If any of the zones or records is invalid, nothing will be replaced.

        This method is useful for reload configuration from disk
        or database or etc.
        """
        new_zones: RadixTree[DNSZone] = RadixTree()
        for zone_name, records in zones_data.items():
            zone = DNSZone(zone_name, *records)
            new_zones.insert(self.get_reverse_tuple(zone.name), zone)
        self.zones = new_zones



import asyncio
import logging
import os
import socket
from typing import (
    Any, Iterator, Optional, ParamSpec, Protocol, TypeAlias, final,
)

from ._context_vars import EVENT_LOOP


log = logging.getLogger(__name__)


try:
    from time import time_ns
except ImportError:
    from time import time

    def time_ns() -> int:
        return int(time() * 1000000000)


try:
    from typing import Concatenate
except ImportError:
    from typing_extensions import Concatenate


class EntrypointProtocol(Protocol):
    @property
    def name(self) -> str:
        ...

    def load(self) -> Any:
        ...


# noinspection PyUnresolvedReferences
try:
    from importlib.metadata import Distribution, EntryPoint

    def entry_pont_iterator(entry_point: str) -> Iterator[EntrypointProtocol]:
        ep: EntryPoint
        for dist in Distribution.discover():
            for ep in dist.entry_points:
                if ep.group == entry_point:
                    yield ep
except ImportError:
    import pkg_resources

    def entry_pont_iterator(entry_point: str) -> Iterator[EntrypointProtocol]:
        yield from pkg_resources.iter_entry_points(entry_point)


class EventLoopMixin:
    __slots__ = "_loop",

    _loop: Optional[asyncio.AbstractEventLoop]

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        if not getattr(self, "_loop", None):
            self._loop = asyncio.get_running_loop()
        return self._loop  # type: ignore


event_loop_policy: asyncio.AbstractEventLoopPolicy
try:
    import uvloop

    if (
        os.getenv("AIOMISC_USE_UVLOOP", "1").lower() in
        ("yes", "1", "enabled", "enable", "on", "true")
    ):
        event_loop_policy = uvloop.EventLoopPolicy()
    else:
        event_loop_policy = asyncio.DefaultEventLoopPolicy()
except ImportError:
    event_loop_policy = asyncio.DefaultEventLoopPolicy()

if hasattr(socket, "TCP_NODELAY"):
    def sock_set_nodelay(sock: socket.socket) -> None:
        if sock.proto != socket.IPPROTO_TCP:
            return None
        sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
else:
    def sock_set_nodelay(sock: socket.socket) -> None:
        return None

if hasattr(socket, "SO_REUSEPORT"):
    def sock_set_reuseport(sock: socket.socket, reuse_port: bool) -> None:
        sock.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEPORT, reuse_port.real,
        )
else:
    def sock_set_reuseport(sock: socket.socket, reuse_port: bool) -> None:
        log.debug(
            "SO_REUSEPORT is not implemented by "
            "underlying library. Skipping.",
        )

# 16.x.x reverse compatibility
set_current_loop = EVENT_LOOP.set
get_current_loop = EVENT_LOOP.get

__all__ = (
    "Concatenate",
    "EntrypointProtocol",
    "EventLoopMixin",
    "ParamSpec",
    "Protocol",
    "TypeAlias",
    "entry_pont_iterator",
    "event_loop_policy",
    "final",
    "get_current_loop",
    "set_current_loop",
    "sock_set_nodelay",
    "sock_set_reuseport",
    "time_ns",
)



from collections import defaultdict
from typing import DefaultDict, Iterable, Sequence, Set, Tuple

from .records import DNSRecord, RecordType


RecordsType = DefaultDict[Tuple[str, RecordType], Set[DNSRecord]]


class DNSZone:
    records: RecordsType
    name: str

    __slots__ = ("name", "records")

    def __init__(self, name: str, *records: DNSRecord) -> None:
        if not name.endswith("."):
            name += "."
        self.name = name
        self.records = defaultdict(set)

        for record in records:
            self.add_record(record)

    def add_record(self, record: DNSRecord) -> None:
        if not self.check_record(record):
            raise ValueError(
                f"Record {record.name} does not belong to zone {self.name}",
            )
        self.records[(record.name, record.type)].add(record)

    def remove_record(self, record: DNSRecord) -> None:
        key = (record.name, record.type)
        if key in self.records:
            self.records[key].discard(record)
            if self.records[key]:
                return
            del self.records[key]

    def get_records(
        self, name: str, record_type: RecordType,
    ) -> Sequence[DNSRecord]:
        if not name.endswith("."):
            name += "."
        return tuple(self.records.get((name, record_type), ()))

    def check_record(self, record: DNSRecord) -> bool:
        return record.name.endswith(self.name)

    def replace(self, records: Iterable[DNSRecord]) -> None:
        """
        Atomically replace all records in specified zone with new ones.
        This method is safe because it replaces all records at once.

        If any of the records does not belong to the zone, ValueError
        will be raised and no records will be replaced.
        """
        new_records: RecordsType = defaultdict(set)

        for record in records:
            if not self.check_record(record):
                raise ValueError(
                    f"Record {record.name} does not "
                    f"belong to zone {self.name}",
                )
            new_records[(record.name, record.type)].add(record)
        self.records = new_records



import asyncio
import socket
import ssl
from contextlib import suppress
from typing import Any, Dict, Iterable, Mapping, Optional, Tuple, Union

from aiohttp.web import Application, AppRunner, BaseRunner, SockSite  # noqa

from aiomisc.service.tls import PathOrStr, SSLOptions

from ..utils import bind_socket
from .base import Service


try:
    from aiohttp.web_log import AccessLogger
except ImportError:         # pragma: nocover
    from aiohttp.helpers import AccessLogger  # type: ignore


RunnerKwargsType = Union[Mapping[str, Any], Iterable[Tuple[str, Any]]]


class AIOHTTPService(Service):
    __async_required__: Tuple[str, ...] = (
        "start", "create_application",
    )

    site: SockSite
    runner: BaseRunner
    handler_cancellation: bool = True

    def __init__(
        self, address: Optional[str] = "localhost", port: Optional[int] = None,
        sock: Optional[socket.socket] = None, shutdown_timeout: int = 5,
        handler_cancellation: bool = handler_cancellation,
        runner_kwargs: Optional[RunnerKwargsType] = None,
        **kwds: Any,
    ):

        if not sock:
            if not (address and port):
                raise RuntimeError(
                    "You should pass socket instance or "
                    '"address" and "port" couple',
                )

            self.socket = bind_socket(
                address=address,
                port=port,
                proto_name="http",
            )

        elif not isinstance(sock, socket.socket):
            raise ValueError("sock must be socket instance")
        else:
            self.socket = sock

        self.shutdown_timeout = shutdown_timeout

        self.runner_kwargs: Dict[str, Any] = dict(runner_kwargs or {})
        self.runner_kwargs.setdefault("access_log_class", AccessLogger)
        self.runner_kwargs.setdefault(
            "access_log_format", AccessLogger.LOG_FORMAT,
        )
        self.runner_kwargs.setdefault(
            "handler_cancellation", handler_cancellation,
        )

        super().__init__(**kwds)

    async def create_application(self) -> Application:
        return Application()

    async def create_site(self) -> SockSite:
        if getattr(self, "runner", None) is None:
            raise RuntimeError("runner already created")

        return SockSite(self.runner, self.socket)

    async def make_runner(self, application: Application) -> AppRunner:
        return AppRunner(
            application,
            shutdown_timeout=self.shutdown_timeout,
            **self.runner_kwargs,
        )

    async def start(self) -> None:
        if hasattr(self, "runner"):
            raise RuntimeError("Can not start twice")

        application: Application = await self.create_application()
        self.runner = await self.make_runner(application)
        await self.runner.setup()

        self.site = await self.create_site()
        await self.site.start()

    async def stop(self, exception: Optional[Exception] = None) -> None:
        try:
            if self.site:
                await self.site.stop()
        finally:
            # Ensure all handlers are finished
            await asyncio.sleep(0)
            if hasattr(self, "runner"):
                # Avoid AttributeError in case server already dereferenced
                with suppress(AttributeError):
                    await self.runner.cleanup()


class AIOHTTPSSLService(AIOHTTPService):
    def __init__(
        self, cert: PathOrStr, key: PathOrStr, ca: Optional[PathOrStr] = None,
        address: Optional[str] = None, port: Optional[int] = None,
        verify: bool = True, sock: Optional[socket.socket] = None,
        shutdown_timeout: int = 5, require_client_cert: bool = False,
        **kwds: Any,
    ):

        super().__init__(
            address=address, port=port, sock=sock,
            shutdown_timeout=shutdown_timeout, **kwds,
        )

        self.__ssl_options = SSLOptions(
            cert, key, ca, verify, require_client_cert,
            ssl.Purpose.CLIENT_AUTH,
        )

    async def create_site(self) -> SockSite:
        assert self.runner and self.socket

        return SockSite(
            self.runner, self.socket,
            ssl_context=await self.loop.run_in_executor(
                None, self.__ssl_options.create_context,
            ),
        )



import asyncio
from functools import wraps
from typing import (
    Any, Callable, Coroutine, Generic, Optional, ParamSpec, Tuple, Type,
    TypeVar, Union,
)

from .counters import Statistic


Number = Union[int, float]
T = TypeVar("T")
P = ParamSpec("P")


class BackoffStatistic(Statistic):
    done: int
    attempts: int
    cancels: int
    errors: int
    sum_time: float


class RetryStatistic(BackoffStatistic):
    pass


class Backoff:
    __slots__ = (
        "attempt_timeout",
        "deadline",
        "pause",
        "max_tries",
        "giveup",
        "exceptions",
        "statistic",
    )

    def __init__(
        self,
        attempt_timeout: Optional[Number],
        deadline: Optional[Number],
        pause: Number = 0,
        exceptions: Tuple[Type[Exception], ...] = (),
        max_tries: Optional[int] = None,
        giveup: Optional[Callable[[Exception], bool]] = None,
        statistic_name: Optional[str] = None,
        statistic_class: Type[BackoffStatistic] = BackoffStatistic,
    ):
        if not pause:
            pause = 0
        elif pause < 0:
            raise ValueError("'pause' must be positive")

        if attempt_timeout is not None and attempt_timeout < 0:
            raise ValueError("'attempt_timeout' must be positive or None")

        if deadline is not None and deadline < 0:
            raise ValueError("'deadline' must be positive or None")

        if max_tries is not None and max_tries < 1:
            raise ValueError("'max_retries' must be >= 1 or None")

        if giveup is not None and not callable(giveup):
            raise ValueError("'giveup' must be a callable or None")

        exceptions = tuple(exceptions) or ()
        exceptions += asyncio.TimeoutError,

        self.attempt_timeout = attempt_timeout
        self.deadline = deadline
        self.pause = pause
        self.max_tries = max_tries
        self.giveup = giveup
        self.exceptions = exceptions
        self.statistic = statistic_class(statistic_name)

    def prepare(
        self,
        func: Callable[P, Coroutine[Any, Any, T]],
    ) -> "BackoffExecution[P, T]":
        return BackoffExecution(
            function=func,
            statistic=self.statistic,
            attempt_timeout=self.attempt_timeout,
            deadline=self.deadline,
            pause=self.pause,
            max_tries=self.max_tries,
            giveup=self.giveup,
            exceptions=self.exceptions,
        )

    async def execute(
        self,
        func: Callable[P, Coroutine[Any, Any, T]],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> T:
        execution = self.prepare(func)
        return await execution(*args, **kwargs)

    def __call__(
        self,
        func: Callable[P, Coroutine[Any, Any, T]],
    ) -> Callable[P, Coroutine[Any, Any, T]]:
        if not asyncio.iscoroutinefunction(func):
            raise TypeError("Function must be a coroutine function")

        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            return await self.execute(func, *args, **kwargs)

        return wrapper


class BackoffExecution(Generic[P, T]):
    __slots__ = (
        "attempt_timeout",
        "deadline",
        "exceptions",
        "function",
        "giveup",
        "last_exception",
        "max_tries",
        "pause",
        "statistic",
        "total_tries",
    )

    def __init__(
        self,
        function: Callable[P, Coroutine[Any, Any, T]],
        statistic: BackoffStatistic,
        attempt_timeout: Optional[Number],
        deadline: Optional[Number],
        pause: Number = 0,
        exceptions: Tuple[Type[Exception], ...] = (),
        max_tries: Optional[int] = None,
        giveup: Optional[Callable[[Exception], bool]] = None,
    ):
        self.function = function
        self.statistic = statistic
        self.attempt_timeout = attempt_timeout
        self.deadline = deadline
        self.pause = pause
        self.max_tries = max_tries
        self.giveup = giveup
        self.exceptions = exceptions

        self.last_exception: Optional[Exception] = None
        self.total_tries: int = 0

    async def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        return await self.execute(*args, **kwargs)

    async def execute(self, *args: P.args, **kwargs: P.kwargs) -> T:
        async def run() -> Any:
            loop = asyncio.get_running_loop()

            while True:
                self.statistic.attempts += 1
                self.total_tries += 1
                delta = -loop.time()

                try:
                    return await asyncio.wait_for(
                        self.function(*args, **kwargs),
                        timeout=self.attempt_timeout,
                    )
                except asyncio.CancelledError:
                    self.statistic.cancels += 1
                    raise
                except self.exceptions as e:
                    self.statistic.errors += 1
                    self.last_exception = e
                    if (
                        self.max_tries is not None and
                        self.total_tries >= self.max_tries
                    ):
                        raise

                    if self.giveup and self.giveup(e):
                        raise
                    await asyncio.sleep(self.pause)
                except Exception as e:
                    self.last_exception = e
                    raise
                finally:
                    delta += loop.time()
                    self.statistic.sum_time += delta
                    self.statistic.done += 1

        try:
            return await asyncio.wait_for(run(), timeout=self.deadline)
        except Exception:
            if self.last_exception is not None:
                raise self.last_exception
            raise


# noinspection SpellCheckingInspection
def asyncbackoff(
    attempt_timeout: Optional[Number],
    deadline: Optional[Number],
    pause: Number = 0,
    *exc: Type[Exception], exceptions: Tuple[Type[Exception], ...] = (),
    max_tries: Optional[int] = None,
    giveup: Optional[Callable[[Exception], bool]] = None,
    statistic_name: Optional[str] = None,
    statistic_class: Type[BackoffStatistic] = BackoffStatistic,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, T]]],
    Callable[P, Coroutine[Any, Any, T]],
]:
    """
    Patametric decorator that ensures that ``attempt_timeout`` and
    ``deadline`` time limits are met by decorated function.

    In case of exception function will be called again with similar
    arguments after ``pause`` seconds.

    :param statistic_name: name filed for statistic instances
    :param attempt_timeout: is maximum execution time for one
                            execution attempt.
    :param deadline: is maximum execution time for all execution attempts.
    :param pause: is time gap between execution attempts.
    :param exc: retrying when these exceptions were raised.
    :param exceptions: similar as exc but keyword only.
    :param max_tries: is maximum count of execution attempts (>= 1).
    :param giveup: is a predicate function which can decide by a given
    :param statistic_class: statistic class
    """

    return Backoff(
        attempt_timeout=attempt_timeout,
        deadline=deadline,
        pause=pause,
        exceptions=exceptions or exc,
        max_tries=max_tries,
        giveup=giveup,
        statistic_name=statistic_name,
        statistic_class=statistic_class,
    )


def asyncretry(
    max_tries: Optional[int],
    exceptions: Tuple[Type[Exception], ...] = (Exception,),
    pause: Number = 0,
    giveup: Optional[Callable[[Exception], bool]] = None,
    statistic_name: Optional[str] = None,
) -> Callable[
    [Callable[P, Coroutine[Any, Any, T]]],
    Callable[P, Coroutine[Any, Any, T]],
]:
    """
    Shortcut of ``asyncbackoff(None, None, 0, Exception)``.

    In case of exception function will be called again with similar
    arguments after ``pause`` seconds.

    :param max_tries: is maximum count of execution attempts
                      (>= 1 or ``None`` means infinity).
    :param exceptions: similar as exc but keyword only.
    :param giveup: is a predicate function which can decide by a given
    :param pause: is time gap between execution attempts.
    :param statistic_name: name filed for statistic instances
    """

    return asyncbackoff(
        attempt_timeout=None,
        deadline=None,
        exceptions=exceptions,
        giveup=giveup,
        max_tries=max_tries,
        pause=pause,
        statistic_class=RetryStatistic,
        statistic_name=statistic_name,
    )
